from datetime import datetime, timedelta
from typing import Generator, Optional

from fastapi import Depends, FastAPI, Request, Response
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from sqlalchemy import and_, or_
from sqlalchemy.orm import Session

from bebe import db


app = FastAPI()
app.mount("/static", StaticFiles(directory="static"), name="static")
templates = Jinja2Templates(directory="templates")


def form_to_datetime(time: str, fallback: Optional[datetime]) -> Optional[datetime]:
    now = datetime.now()
    try:
        (hour, minute) = (int(t) for t in time.split(":"))
        return datetime(
            year=now.year, month=now.month, day=now.day, hour=hour, minute=minute
        )
    except Exception as _:
        return fallback


def dbsession() -> Generator[Session, None, None]:
    session = db.SessionLocal()
    try:
        yield session
    finally:
        session.close()


def today() -> str:
    return datetime.today().strftime("%Y-%m-%d")


def timeline(
    request: Request, session: Session, template: str, date: Optional[datetime] = None
) -> Response:
    if date is None:
        date = datetime.today()
    date = datetime(year=date.year, month=date.month, day=date.day)
    timeline = (
        session.query(db.Event)
        .filter(
            and_(
                db.Event.start_time >= date,
                or_(
                    db.Event.end_time.is_(None),
                    db.Event.end_time < date + timedelta(days=1),
                ),
            )
        )
        .order_by(db.Event.start_time.asc())
        .all()
    )
    vkupno_izedeno = sum(
        int(item.data["value"])
        for item in timeline
        if item.event_type == db.EventTypes.hranenje
    )

    return templates.TemplateResponse(
        request=request,
        name=template,
        context={
            "timeline": timeline,
            "date": date.strftime("%Y-%m-%d"),
            "today": datetime.today().strftime("%Y-%m-%d"),
            "vkupno_izedeno": vkupno_izedeno,
        },
    )


@app.get("/")
async def index(request: Request, session: Session = Depends(dbsession)) -> Response:
    return timeline(request, session, "base.html")


@app.get("/formi")
async def index(request: Request, session: Session = Depends(dbsession)) -> Response:
    return timeline(request, session, "form.html")


@app.get("/statistika")
async def statistika(
    request: Request, session: Session = Depends(dbsession)
) -> Response:
    return templates.TemplateResponse(request=request, name="stats.html")


@app.get("/istorija")
@app.post("/istorija")
async def istorija(
    request: Request,
    session: Session = Depends(dbsession),
    den: Optional[str] = None,
) -> Response:
    den = today() if den is None else den
    return timeline(
        request, session, "istorija.html", datetime.strptime(den, "%Y-%m-%d")
    )


@app.post("/event")
async def add_event(
    request: Request, session: Session = Depends(dbsession)
) -> Response:
    data = await request.form()

    event_type = data.get("event_type")
    start_time = form_to_datetime(
        data.get(f"{event_type}_start"), datetime.now().replace(second=0, microsecond=0)
    )
    end_time = form_to_datetime(data.get(f"{event_type}_end"), None)

    event = db.Event(
        start_time=start_time,
        end_time=end_time,
        event_type=event_type,
    )

    value = data.get(f"{event_type}_data_value")
    if value is not None:
        event.data = {
            "value": value,
            "unit": data.get(f"{event_type}_data_unit"),
        }
    session.add(event)
    session.commit()

    return timeline(request, session, "timeline.html")


@app.delete("/event/{event_id}")
async def remove_event(
    request: Request, event_id: int, session: Session = Depends(dbsession)
) -> Response:
    event = session.get(db.Event, event_id)
    session.delete(event)
    session.commit()

    return timeline(request, session, "timeline.html")
